#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#
import abc
from collections import defaultdict
import uuid

from asgiref.sync import sync_to_async

from django_quotas.base.dto import Quota, QuotaBucket, QuotaStats, QuotaUseForBucket, ValuePerBucket

__all__ = [
    "QuotaExceededError",
    "QuotaService",
]

from django_quotas.base.features import FeatureId


class QuotaExceededError(Exception):
    """Exception raised when a quota is exceeded for an account and feature(s).

    :param account_name: The account UUID.
    :param exceeded_features: Mapping of feature names to exceeded bucket stats.
    """

    def __init__(self, account_name: str, exceeded_features: dict[str, list[QuotaUseForBucket]]):
        self.account_id = account_name
        self._exceeded_features_stats = exceeded_features
        super().__init__(self.__generate_detailed_message())

    def get_exceeded_features(self) -> set[str]:
        """Get the set of feature names for which quotas are exceeded.

        :return: Set of feature names.
        """
        return set(self._exceeded_features_stats.keys())

    def get_stats_per_feature(self) -> dict[str, list[QuotaUseForBucket]]:
        """Get the exceeded stats per feature.

        :return: Mapping of feature names to list of QuotaUseForBucket.
        """
        return self._exceeded_features_stats

    def __generate_detailed_message(self) -> str:
        msg = self.__generate_msg() + "\nDetails:"
        for feature, stats in self._exceeded_features_stats.items():
            msg += f"\n{feature}: "
            buckets = []
            for stat in stats:
                buckets.append(f"{stat.bucket_name}({stat.current_usage}/{stat.limit})")
            msg += ", ".join(buckets)
        return msg

    def __generate_msg(self) -> str:
        return f"Quota exceeded for account {self.account_id}. Exceeded features: {self.get_exceeded_features()}"


class QuotaService(metaclass=abc.ABCMeta):
    """Abstract base class for quota service implementations."""

    def ensure_quota_or_raise(
        self, account_id: str, feature: FeatureId | set[FeatureId], potential_increase: int = 1
    ) -> None:
        """
        Ensure that the quota for the given account and feature is not exceeded.

        If the quota is exceeded, raise a QuotaExceededError.
        :param account_id: The account ID.
        :param feature: The feature name or set of feature names.
        :param potential_increase: The potential increase in usage.\
            Current utilization + potential increase must be less than the quota.
        :raise QuotaExceededError: If the quota is exceeded.
        """
        feature_names: set[str] = self.resolve_feature_names(feature)
        utilization = self.get_quotas_utilization(account_id, feature)
        exceeded_features: dict[str, list[QuotaUseForBucket]] = defaultdict(list)
        for feature in feature_names:
            if feature not in utilization.feature_stats:
                continue  # this means we have no quota for this specific feature
            status = utilization.feature_stats[feature]
            quota_buckets = [
                (QuotaBucket.HOURLY, status.usage.hourly or 0, status.limits.hourly),
                (QuotaBucket.DAILY, status.usage.daily or 0, status.limits.daily),
                (QuotaBucket.MONTHLY, status.usage.monthly or 0, status.limits.monthly),
                (QuotaBucket.TOTAL, status.usage.total or 0, status.limits.total),
            ]

            for bucket_name, current_usage, limit in quota_buckets:
                if limit is not None and current_usage + potential_increase > limit:
                    exceeded_features[feature].append(
                        QuotaUseForBucket(
                            bucket_name=bucket_name,
                            current_usage=current_usage,
                            limit=limit,
                            quota_id=status.quota_id,
                        )
                    )
        if exceeded_features:
            raise QuotaExceededError(account_id, exceeded_features)

    async def aensure_quota_or_raise(
        self, account_id: str, feature: FeatureId | set[FeatureId], potential_increase: int = 1
    ) -> None:
        """
        Ensure that the quota for the given account and feature is not exceeded.

        If the quota is exceeded, raise a QuotaExceededError.
        :param account_id: The account ID.
        :param feature: The feature name.
        :param potential_increase: The potential increase in usage.\
            Current utilization + potential_increase must be less than the quota.
        """
        return await sync_to_async(self.ensure_quota_or_raise)(account_id, feature, potential_increase)

    @abc.abstractmethod
    def register_usage(self, account_id: str, feature: FeatureId, increment: int = 1) -> None:
        """
        Register usage for the given account and feature.

        :param account_id: The account ID.
        :param feature: The feature name.
        :param increment: The value to be added to quota usage.
        """
        pass

    @abc.abstractmethod
    async def aregister_usage(self, account_id: str, feature: FeatureId | set[FeatureId], increment: int = 1) -> None:
        """
        Register usage for the given account and feature.

        :param account_id: The account ID.
        :param feature: The feature name.
        :param increment: The value to be added to quota usage.
        """
        pass

    @abc.abstractmethod
    def get_quotas_utilization(self, account_id: str, feature: FeatureId | set[FeatureId] | None) -> QuotaStats:
        """
        Get the quota utilization for the given account and features.

        :param account_id: The account ID.
        :param feature: Name of the feature or a list of feature names. None means all features having quotas.
        """
        pass

    def _get_account_id_for_quota_search(self, account_id: str) -> str:
        """
        Get the account ID to be used for quota search.
        This method can be overridden in subclasses to modify the account ID e.g. if for anonymous users account ID
        is generated dynamically e.g. based on IP address or session ID you most likely want to setup quota for ANY
        anonymous user rather than for each unique anonymous user. So this is the right place to transform account ID
        into a common one for all anonymous users.

        :param account_id: The account ID.
        :return: The account ID to be used for quota search.
        """
        return account_id

    @abc.abstractmethod
    async def aget_quotas_utilization(self, account_id: str, feature: FeatureId | set[FeatureId] | None) -> QuotaStats:
        """
        Get the quota utilization for the given account and features.

        :param account_id: The account ID.
        :param feature: Name of the feature or a list of feature names. None means all features having quotas.
        """
        pass

    @abc.abstractmethod
    def set_quota(
        self, account_id: str, feature: FeatureId, limits: ValuePerBucket, owner_tag: str | None = None
    ) -> Quota:
        """
        Set the quota for the given account and feature.

        :param account_id: The account ID.
        :param feature: The feature name.
        :param limits: Limits for each bucket.
        :param owner_tag: Optional owner tag.
        :return: The created or updated quota.
        """
        pass

    @abc.abstractmethod
    async def aset_quota(
        self, account_id: str, feature: FeatureId, limits: ValuePerBucket, owner_tag: str | None = None
    ) -> Quota:
        """
        Set the quota for the given account and feature.

        :param account_id: The account ID.
        :param feature: The feature name.
        :param limits: Limits for each bucket.
        :param owner_tag: Optional owner tag.
        :return: The created or updated quota.
        """
        pass

    @staticmethod
    def resolve_feature_names(features: FeatureId | set[FeatureId]) -> set[str]:
        """
        Resolve feature names from FeatureId or set of FeatureId.

        :param features: FeatureId or set of FeatureId.
        :return: Set of feature names.
        """
        if isinstance(features, set):
            return {QuotaService.resolve_feature_name(f) for f in features}
        return {QuotaService.resolve_feature_name(features)}

    @staticmethod
    def resolve_feature_name(feature: FeatureId) -> str:
        """
        Resolve feature name from FeatureId.

        :param feature: FeatureId.
        :return: Feature name.
        """
        if isinstance(feature, str):
            return feature
        return feature.full_name



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#
__all__ = ("DefaultQuotaModel",)

from typing import cast
import uuid

from django.db import models
from django.utils.translation import gettext_lazy as gt

from django_quotas.config import DjangoQuotasConfig as cfg
from django_quotas.models import BaseQuotaModel, QuotaModelMetaclass


class DefaultQuotaModel(BaseQuotaModel, metaclass=QuotaModelMetaclass):
    """Model for actual quota assigned to a user."""

    class Meta:
        db_table = f"{cfg.TABLE_PREFIX}_quota"
        verbose_name = gt("Account Quota")
        unique_together = ("account", "feature_name")
        indexes = (models.Index(fields=["account", "feature_name"]),)

    account: models.Model = models.ForeignKey(  # type: ignore[assignment,call-arg]
        cfg.QUOTA_RELATED_ACCOUNT_MODEL,
        null=False,
        blank=False,
        on_delete=models.CASCADE,
        related_name="quotas",
        db_index=True,
        swappable=True,
    )

    def __str__(self) -> str:
        return f"Quota(id={self.pk}, feature={self.feature_name})"



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#
import datetime
from typing import cast

from django.apps import apps
from django.db import models
from django.utils.module_loading import import_string

__all__ = ["get_model_by_name", "get_class_by_name", "datetime_now"]


def get_model_by_name(model_name: str) -> type[models.Model]:
    """Get a Django model class by its dotted name.

    :param model_name: Dotted model name (e.g., 'app_label.ModelName').
    :return: Model class.
    :raises ValueError: If the model cannot be found.
    """
    try:
        app_label, model = model_name.split(".", 1)
        return cast(type[models.Model], apps.get_model(app_label, model))
    except LookupError as err:
        raise ValueError(f"Model {model_name} not found") from err


def get_class_by_name(dotted_path: str) -> type:
    """Import and return a class by its dotted Python path.

    :param dotted_path: Dotted Python path to the class.
    :return: Imported class type.
    """
    return import_string(dotted_path)


def datetime_now() -> datetime.datetime:
    """Return the current datetime in UTC.

    :return: Current datetime with UTC timezone.
    """
    return datetime.datetime.now(tz=datetime.UTC)



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#
import asyncio
import datetime
from typing import Any, cast

from asgiref.sync import sync_to_async
from django.db import transaction
from django.db.models import F, Sum

from django_quotas.backends.db.base_models import BaseQuotaUsageModel
from django_quotas.backends.db.config import DjangoQuotasDbConfig as cfg_db
from django_quotas.base.dto import Quota, QuotaStats, QuotaStatus, ValuePerBucket
from django_quotas.base.features import FeatureId
from django_quotas.base.service import QuotaService
from django_quotas.config import DjangoQuotasConfig as cfg
from django_quotas.models import BaseQuotaModel
from django_quotas.utils import datetime_now


class DbQuotaService(QuotaService):
    def __init__(
        self,
        usage_model_cls: type[BaseQuotaUsageModel] = cfg_db.quota_usage_cls,
        quota_model_cls: type[BaseQuotaModel] = cfg.quota_cls,
    ) -> None:
        self.usage_model_cls = usage_model_cls
        self.quota_model_cls = quota_model_cls

    def register_usage(self, account_id: str, feature: FeatureId, increment: int = 1) -> None:
        """Register usage for the given account and feature.

        :param account_id: The account ID.
        :param feature: The feature name.
        :param increment: The value to be added to quota usage.
        """
        feature_name = self.resolve_feature_name(feature)
        # Current hour
        current_time = datetime_now().replace(minute=0, second=0, microsecond=0)

        with transaction.atomic():
            usage, _ = self.usage_model_cls.objects.get_or_create(  # type: ignore[attr-defined]
                account_id=account_id, feature_name=feature_name, point_in_time=current_time
            )
            self.usage_model_cls.objects.filter(id=usage.id, point_in_time=current_time).update(  # type: ignore[attr-defined]
                usage_count=F("usage_count") + increment
            )

    async def aregister_usage(self, account_id: str, feature: FeatureId | set[FeatureId], increment: int = 1) -> None:
        """Asynchronously register usage for the given account and feature(s).

        :param account_id: The account ID.
        :param feature: The feature name or set of feature names.
        :param increment: The value to be added to quota usage.
        """
        feature_names = self.resolve_feature_names(feature)
        await asyncio.gather(*[sync_to_async(self.register_usage)(account_id, f, increment) for f in feature_names])

    def get_quotas_utilization(self, account_id: str, feature: FeatureId | set[FeatureId] | None) -> QuotaStats:
        """Get the quota utilization for the given account and features.

        :param account_id: The account ID.
        :param feature: Name of the feature or a set of feature names. None means all features having quotas.
        :return: QuotaStats instance with utilization data.
        """
        feature_names = self.resolve_feature_names(feature) if feature is not None else None
        current_hour = self.__get_current_hour()  # Current hour
        today = current_hour.replace(hour=0, minute=0, second=0, microsecond=0)  # Current day
        first__day_of_month = today.replace(day=1)
        quotas_qs = self.quota_model_cls.objects.filter(account_id=self._get_account_id_for_quota_search(account_id))  # type: ignore[attr-defined]
        if feature_names:
            quotas_qs = quotas_qs.filter(feature_name__in=feature_names)

        quotas: list[BaseQuotaModel] = list(quotas_qs)
        feature_names = {quota.feature_name for quota in quotas}

        # Calculate usage per bucket

        hourly_usage = self.__create_usage_for_bucket(account_id, feature_names, point_in_time=current_hour)
        daily_usage = self.__create_usage_for_bucket(account_id, feature_names, point_in_time__date=today)
        monthly_usage = self.__create_usage_for_bucket(
            account_id, feature_names, point_in_time__date__gte=first__day_of_month
        )
        total_usage = self.__create_usage_for_bucket(account_id, feature_names)

        result = QuotaStats(account_id=account_id, feature_stats={})

        # Compile the stats for each feature
        for quota in quotas:
            result.feature_stats[quota.feature_name] = QuotaStatus(
                quota_id=quota.id,
                limits=ValuePerBucket(
                    hourly=quota.hourly_limit,
                    daily=quota.daily_limit,
                    monthly=quota.monthly_limit,
                    total=quota.total_limit,
                ),
                usage=ValuePerBucket(
                    hourly=hourly_usage.get(quota.feature_name, 0),
                    daily=daily_usage.get(quota.feature_name, 0),
                    monthly=monthly_usage.get(quota.feature_name, 0),
                    total=total_usage.get(quota.feature_name, 0),
                ),
            )

        return result

    async def aget_quotas_utilization(self, account_id: str, feature: FeatureId | set[FeatureId] | None) -> QuotaStats:
        return await sync_to_async(self.get_quotas_utilization)(account_id, feature)

    def set_quota(
        self, account_id: str | int, feature: FeatureId, limits: ValuePerBucket, owner_tag: str | None = None
    ) -> Quota:
        quota_model, _ = self.quota_model_cls.objects.update_or_create(  # type: ignore[attr-defined]
            account_id=account_id,
            feature_name=self.resolve_feature_name(feature),
            defaults={
                "hourly_limit": limits.hourly,
                "daily_limit": limits.daily,
                "monthly_limit": limits.monthly,
                "total_limit": limits.total,
                "owner_tag": owner_tag,
            },
        )
        return quota_model

    async def aset_quota(
        self, account_id: str, feature: FeatureId, limits: ValuePerBucket, owner_tag: str | None = None
    ) -> Quota:
        return await sync_to_async(self.set_quota)(account_id, feature, limits, owner_tag)

    @staticmethod
    def __get_current_hour() -> datetime.datetime:
        return datetime_now().replace(minute=0, second=0, microsecond=0)

    def __create_usage_for_bucket(
        self, account_id: str, feature_name: set[str] | None, **extra_query_args: Any
    ) -> dict[str, int]:
        """
        Get usage for the given account and, optionally, list of specific features.

        Returns a dict with feature names as keys and usage as values.
        """
        qs = self.usage_model_cls.objects.filter(  # type: ignore[attr-defined]
            account_id=cast(Any, account_id), feature_name__in=feature_name, **extra_query_args
        )
        if feature_name is not None:
            qs = qs.filter(feature_name__in=feature_name)
        qs_annotated = qs.values("feature_name").annotate(total_usage=Sum("usage_count"))
        return {item["feature_name"]: item["total_usage"] for item in qs_annotated}



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#
import uuid

from django.db import models
from django.utils.translation import gettext_lazy as gt

from django_quotas.config import DjangoQuotasConfig as cfg


class BaseQuotaUsageModel(models.Model):
    class Meta:
        abstract = True
        db_table = f"{cfg.TABLE_PREFIX}_quota_usage"
        verbose_name = gt("Quota Usage")
        unique_together = ("account_id", "feature_name", "point_in_time")
        indexes = (
            models.Index(fields=["account_id", "feature_name", "point_in_time"]),
            models.Index(fields=["account_id", "point_in_time"]),
        )

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    account_id = models.CharField(verbose_name="Account", null=False, blank=False, db_index=True)
    feature_name = models.CharField(max_length=300, null=False, blank=False)
    point_in_time = models.DateTimeField(null=False, blank=False)
    usage_count = models.IntegerField(null=False, blank=False, default=0)

    def __str__(self) -> str:
        return f"QuotaUsage(account={self.account_id}, feature={self.feature_name}, ts={self.point_in_time})"



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#

from django_quotas.backends.db.base_models import BaseQuotaUsageModel


class QuotaUsageModel(BaseQuotaUsageModel):
    """Model for tracking quota usage per account, feature, and time point."""

    pass



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#

__all__ = ["BaseQuotaModel", "QuotaModelMetaclass"]

import abc
import uuid

from django.db import models

from django_quotas.base.dto import Quota, ValuePerBucket


class BaseQuotaModel(models.Model):
    """Base model for quotas that can be converted to Quota interface when needed."""

    class Meta:
        abstract = True

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    owner_tag = models.CharField(max_length=300, null=True, blank=True)
    feature_name = models.CharField(max_length=300, null=False, blank=False)
    hourly_limit = models.IntegerField(null=True, blank=True)
    daily_limit = models.IntegerField(null=True, blank=True)
    monthly_limit = models.IntegerField(null=True, blank=True)
    total_limit = models.IntegerField(null=True, blank=True)

    @property
    def get_limits(self) -> ValuePerBucket:
        """Return the quota limits as a ValuePerBucket instance.

        :return: ValuePerBucket with limits for each bucket.
        """
        return ValuePerBucket(
            hourly=self.hourly_limit, daily=self.daily_limit, monthly=self.monthly_limit, total=self.total_limit
        )


class QuotaModelMetaclass(type(models.Model), type(Quota)):  # type: ignore[misc]
    """Specific metaclass to satisfy django migrations creating class in a non-standard way."""

    pass



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#

__all__ = ["QuotasConfig"]

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as gt


class QuotasConfig(AppConfig):
    """Django AppConfig for the django_quotas application."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_quotas"
    label = "django_quotas"
    verbose_name = gt("Quotas")
    default = True



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#
__all__ = ["QuotaFeature", "FeatureSet", "FEATURES", "FeatureId"]
import dataclasses


@dataclasses.dataclass(frozen=True, kw_only=True)
class QuotaFeature:
    namespace: str
    name: str
    verbose_name: str

    @property
    def full_name(self) -> str:
        return f"{self.namespace}.{self.name}" if self.namespace else self.name

    @property
    def id(self) -> str:
        return self.full_name

    def __str__(self) -> str:
        return self.full_name

    def __repr__(self) -> str:
        return f"QuotaFeature(full_name={self.full_name!r}, verbose_name={self.verbose_name!r})"


FeatureId = str | QuotaFeature


class FeatureSet:
    def __init__(self) -> None:
        self._features: dict[str, QuotaFeature] = {}

    def register(self, *feature: QuotaFeature) -> None:
        for f in feature:
            self._features[f.id] = f

    def get(self, feature_id: str) -> QuotaFeature | None:
        return self._features.get(feature_id)

    def all(self) -> list[QuotaFeature]:
        return list(self._features.values())


FEATURES = FeatureSet()



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#

__all__ = ["QuotasDefaultsConfig"]

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as gt


class QuotasDefaultsConfig(AppConfig):
    """Django AppConfig for the django_quotas application."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_quotas.defaults"
    label = "django_quotas_defaults"
    verbose_name = gt("Quotas")
    default = True



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#
import uuid

from django.db import migrations, models
import django.db.models.deletion

import django_quotas.base.dto
from django_quotas.config import DjangoQuotasConfig as cfg


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(cfg.QUOTA_RELATED_ACCOUNT_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="DefaultQuotaModel",
            fields=[
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("owner_tag", models.CharField(blank=True, max_length=300, null=True)),
                ("feature_name", models.CharField(max_length=300)),
                ("hourly_limit", models.IntegerField(blank=True, null=True)),
                ("daily_limit", models.IntegerField(blank=True, null=True)),
                ("monthly_limit", models.IntegerField(blank=True, null=True)),
                ("total_limit", models.IntegerField(blank=True, null=True)),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="quotas",
                        to=cfg.QUOTA_RELATED_ACCOUNT_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Account Quota",
                "db_table": "django_quotas_quota",
                "indexes": [models.Index(fields=["account", "feature_name"], name="django_quot_account_348168_idx")],
                "unique_together": {("account", "feature_name")},
            },
        ),
    ]



#
#  Copyright 2026 by Dmitry Berezovsky, MIT License
#
__all__ = ["QuotasDbConfig"]

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as gt


class QuotasDbConfig(AppConfig):
    """Django AppConfig for the django_quotas.backend.sdb application."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_quotas.backends.db"
    label = "django_quotas_db"
    verbose_name = gt("Quotas DB Backend")
    default = True
